from __future__ import annotations

import platform
import re
import tomllib
from abc import ABC, abstractmethod
from dataclasses import asdict
from pathlib import Path
from typing import TYPE_CHECKING

import tomli_w
from anki.storage import Collection

from .config import AnkiConfig, AnkiNoteConfig, DirectoryConfig, NoteTypeConfig
from .question import Question, Tag

if TYPE_CHECKING:
    from anki.notes import Note


def note_to_question(note_type_configs: list[NoteTypeConfig], note: Note) -> Question:
    """Convert an Anki note to a Question."""
    note_type_config = next(config for config in note_type_configs if config.name == note.note_type()["name"])  # type: ignore[index]
    fields = dict(zip(note.keys(), note.fields, strict=True))
    code_field = fields[note_type_config.code_field.name]
    output_field = fields[note_type_config.output_field.name]
    code = extract_target(note_type_config.code_field.pattern, code_field)
    output = extract_target(note_type_config.output_field.pattern, output_field)
    tags = [tag.removeprefix("snip:") for tag in note.tags if tag.startswith("snip:")]
    note_config = AnkiNoteConfig(tags)
    return Question(
        note.id,
        code,
        note_config.image,
        output,
        note_config.check_output,
        note_config.check_format,
        note_config.output_verbosity,
        note_config.compress,
    )


class Repository(ABC):
    @abstractmethod
    def get(self) -> list[Question]:
        raise NotImplementedError

    @abstractmethod
    def write_output(self, question: Question, output: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def write_code(self, question: Question, code: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def add_tag(self, question: Question, tag: Tag) -> None:
        raise NotImplementedError


class DirectoryRepository(Repository):
    def __init__(self, config: DirectoryConfig, dir: Path):
        self.config = config
        self.dir = dir

    def get(self) -> list[Question]:
        print(f"Looking for questions in directory '{self.dir}'.")

        questions: list[Question] = []
        for dirpath, _, _ in self.dir.walk():
            snippet_paths = list(dirpath.glob("main.*"))
            if len(snippet_paths) == 0:
                continue
            elif len(snippet_paths) > 1:
                raise Exception(f"{dirpath} has multiple snippets")
            (snippet_path,) = snippet_paths
            with open(snippet_path) as f:
                code = f.read()

            output_path = dirpath / "output.txt"
            try:
                with open(output_path) as f:
                    output = f.read()
            except FileNotFoundError:
                # create empty, so we can in effect write, not only check, the output
                output_path.touch()
                output = ""

            try:
                with open(dirpath / "snippet_checker.toml", "rb") as f:
                    question_config = tomllib.load(f)
            except FileNotFoundError:
                question_config = {}
            config = DirectoryConfig(**(asdict(self.config) | question_config))

            questions.append(
                Question(
                    id=snippet_path,
                    code=code,
                    image=config.images[snippet_path.suffix.removeprefix(".")],
                    given_output=output,
                    check_output=config.check_output,
                    check_format=config.check_format,
                    output_verbosity=config.output_verbosity,
                    compress=config.compress,
                    review=config.review,
                )
            )

        return questions

    def write_output(self, question: Question, output: str) -> None:
        assert isinstance(question.id, Path)
        (question.id.parent / "output.txt").write_text(output)

    def write_code(self, question: Question, code: str) -> None:
        assert isinstance(question.id, Path)
        question.id.write_text(code)

    def add_tag(self, question: Question, tag: Tag) -> None:
        """Write a tag to snippet_checker.toml in the question's directory.
        The tag indicates special treatment, e.g. don't check output."""
        assert isinstance(question.id, Path)
        config_path = question.id.parent / "snippet_checker.toml"

        try:
            with open(config_path, "rb") as f:
                config = dict(tomllib.load(f))
        except FileNotFoundError:
            config = {}

        if tag == Tag.REVIEW:
            config["review"] = True
        elif tag == Tag.NO_CHECK_OUTPUT:
            config["check_output"] = False
        elif tag == Tag.NO_CHECK_FORMAT:
            config["check_format"] = False
        elif tag == Tag.NO_COMPRESS:
            config["compress"] = False

        with open(config_path, "wb") as f:
            tomli_w.dump(config, f)


class AnkiRepository(Repository):
    def __init__(self, config: AnkiConfig, tag: str):
        assert (config.profile is not None and config.collection_path is None) or (
            config.profile is None and config.collection_path is not None
        )
        if config.profile is not None:
            system = platform.system()
            if system == "Linux":
                path = Path.home() / f".local/share/Anki2/{config.profile}/collection.anki2"
            else:  # mac
                path = Path.home() / f"Library/Application Support/Anki2/{config.profile}/collection.anki2"
        if config.collection_path is not None:
            path = config.collection_path

        self.config = config
        self.collection = Collection(str(path))
        self.tag = tag

    def get(self) -> list[Question]:
        print(f"Looking for notes tagged '{self.tag}'.")
        note_ids = self.collection.find_notes("")
        notes = [self.collection.get_note(id) for id in note_ids]
        self.notes = [note for note in notes if self.tag in note.tags]
        return [note_to_question(self.config.note_types, note) for note in self.notes]

    def write_output(self, question: Question, output: str) -> None:
        "Replace the question's output field target by the given string."
        assert isinstance(question.id, int)
        note = self.collection.get_note(question.id)  # type: ignore[arg-type]  # TODO: more systematic type conversion
        note_type_config = next(config for config in self.config.note_types if config.name == note.note_type()["name"])  # type: ignore[index]
        index = next(i for i, field_name in enumerate(note.keys()) if field_name == note_type_config.output_field.name)
        current = note.fields[index]
        fixed = replace_target(note_type_config.output_field.pattern, current, escape_html(output))
        note.fields[index] = fixed
        self.collection.update_note(note)

    def write_code(self, question, code: str) -> None:
        "Replace the question's code field target by the given string."
        assert isinstance(question.id, int)
        note = self.collection.get_note(question.id)  # type: ignore[arg-type]  # TODO: more systematic type conversion
        note_type_config = next(config for config in self.config.note_types if config.name == note.note_type()["name"])  # type: ignore[index]
        index = next(i for i, field_name in enumerate(note.keys()) if field_name == note_type_config.code_field.name)
        current = note.fields[index]
        fixed = replace_target(note_type_config.code_field.pattern, current, code)
        note.fields[index] = fixed
        self.collection.update_note(note)

    def add_tag(self, question: Question, tag: Tag) -> None:
        "Write a tag to the given question indicating special treatment, e.g. don't check output."
        note = self.collection.get_note(int(question.id))  # type: ignore[arg-type]  # TODO: more systematic type conversion
        note.tags.append("snip:" + tag.value)  # TODO: avoid scattered hard-coded "snip:"
        self.collection.update_note(note)


def unescape_html(html: str) -> str:
    """
    Replace html tags or entity names with text equivalents.

    Anki treats note fields as html, so some characters in the notes are escaped or replaced by html tags.
    Unescape them here, replacing them with their text equivalents, e.g. &lt; becomes <.

    Some of these escapes probably don't occur in the notes, but it does no harm to include them in case.
    """
    return (
        html.replace("<br>", "\n")
        .replace("&lt;", "<")
        .replace("&gt;", ">")
        .replace("&nbsp;", " ")
        .replace("&amp;", "&")
        .replace("&apos;", "'")
        .replace("&quot;", '"')
    )


def escape_html(plain: str) -> str:
    """
    Anki treats note fields as html, so we need to escape some characters.

    I follow the mdn docs rule of thumb: escape &, then <; anything else is optional.
    Note that I escape less than I unescape: liberal in what you accept, conservative in what you emit.
    """
    return plain.replace("&", "&amp;").replace("<", "&lt;")  # order matters


def extract_target(pattern: re.Pattern, raw: str) -> str:
    "Extract the html-unescaped match for the pattern's 'target' group in `raw`."
    m = pattern.search(raw)
    assert m is not None
    return unescape_html(m.group("target"))


def replace_target(pattern: re.Pattern, current: str, target_repl: str) -> str:
    "Replace the match for the pattern's 'target' group in `current` with html-escaped `target_repl`."
    m = pattern.search(current)
    assert m is not None
    start, end = m.span("target")
    return current[:start] + escape_html(target_repl) + current[end:]



import logging
import sys
from argparse import ArgumentParser
from pathlib import Path

from .check_snippets import check_formatting, check_output
from .config import AnkiConfig, DirectoryConfig, get_anki_config, get_directory_config
from .repository import AnkiRepository, DirectoryRepository, Repository

parser = ArgumentParser()
parser.add_argument("-v", "--verbose", action="store_true", help="Enable debug logging.")
parser.add_argument("--anki", "-a", action="store_true", help="Check anki instead of a directory.")
mode = parser.add_mutually_exclusive_group()
mode.add_argument("-c", "--check", action="store_const", const="check", dest="mode", help="Check and report failures (default).")
mode.add_argument(
    "-i",
    "--interactive",
    action="store_const",
    const="interactive",
    dest="mode",
    help="Prompt user to fix, ignore in future, or leave as is.",
)
mode.add_argument("-f", "--fix", action="store_const", const="fix", dest="mode", help="Automatically fix fixable failures.")
parser.set_defaults(mode="check")
parser.add_argument("command", choices=["output", "format"], help="Command to run.")
parser.add_argument("target", help="Directory or anki tag of the questions to check.")


def app() -> int:
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig()
        logging.getLogger("snippet_checker").setLevel(logging.DEBUG)

    repository: Repository
    config: DirectoryConfig | AnkiConfig
    if args.anki:
        config = get_anki_config()
        repository = AnkiRepository(config, args.target)
    else:
        dir = Path(args.target)
        config = get_directory_config(dir)
        repository = DirectoryRepository(config, dir)

    func = check_output if args.command == "output" else check_formatting
    return func(repository, args.mode)


if __name__ == "__main__":
    sys.exit(app())



import os
import re
import tomllib
from dataclasses import dataclass, field
from pathlib import Path

from .question import Tag


@dataclass
class DirectoryConfig:
    images: dict[str, str] = field(default_factory=dict)
    check_format: bool = True
    check_output: bool = True
    output_verbosity: int = 1
    compress: bool = False
    review: bool = False


def get_directory_config_path(start: Path) -> Path | None:
    current = start.resolve()
    while True:
        candidate = current / "snippet_checker.toml"
        if candidate.exists():
            return candidate
        if current.parent == current:
            return None
        current = current.parent


def get_directory_config(dir: Path) -> DirectoryConfig:
    config_path = get_directory_config_path(dir)
    if config_path is not None:
        with open(config_path, "rb") as f:
            config = DirectoryConfig(**tomllib.load(f))
    else:
        config = DirectoryConfig()

    return config


@dataclass(frozen=True)
class FieldConfig:
    """Information required for extracting the target text from an anki note field."""

    name: str
    pattern: re.Pattern


@dataclass(frozen=True)
class NoteTypeConfig:
    """Information required for checking an anki note type."""

    name: str
    code_field: FieldConfig
    output_field: FieldConfig


@dataclass(frozen=True)
class AnkiConfig:
    """Information required for checking an anki deck."""

    note_types: list[NoteTypeConfig]
    profile: str | None = None
    collection_path: Path | None = None

    def __post_init__(self):
        if self.profile is None and self.collection_path is None:
            raise Exception("Must set either profile or collection path")
        if self.profile is not None and self.collection_path is not None:
            raise Exception("Cannot set both profile and collection path")


def get_anki_config_path() -> Path:
    xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
    xdg_config_file = (Path(xdg_config_home) / "snippet-checker" / "snippet-checker.toml") if xdg_config_home is not None else None
    home_config_file = Path(os.environ["HOME"]) / ".snippet-checker" / "snippet-checker.toml"

    if xdg_config_file and xdg_config_file.exists():
        config_path = xdg_config_file
    elif home_config_file.exists():
        config_path = home_config_file
    else:
        raise Exception("No config found")

    return config_path


def get_anki_config() -> AnkiConfig:
    config_path = get_anki_config_path()
    with open(config_path, "rb") as f:
        raw = tomllib.load(f)

    # extra keys are ignored
    return AnkiConfig(
        profile=raw.get("profile"),
        collection_path=Path(raw["collection_path"]) if "collection_path" in raw else None,
        note_types=[
            NoteTypeConfig(
                name=note_config["note_type"],
                code_field=FieldConfig(
                    name=note_config["code_field"]["name"],
                    pattern=re.compile(note_config["code_field"]["pattern"]),
                ),
                output_field=FieldConfig(
                    name=note_config["output_field"]["name"],
                    pattern=re.compile(note_config["output_field"]["pattern"]),
                ),
            )
            for note_config in raw["notes"]
        ],
    )


class AnkiNoteConfig:
    def __init__(self, tags: list[str]) -> None:
        self.image = next(tag for tag in tags if tag.startswith("image:")).removeprefix("image:")
        self.check_output = Tag.NO_CHECK_OUTPUT.value not in tags
        self.check_format = Tag.NO_CHECK_FORMAT.value not in tags
        try:
            self.output_verbosity = int(next(tag for tag in tags if tag.startswith("output_verbosity:")).removeprefix("output_verbosity:"))
        except StopIteration:
            self.output_verbosity = 0
        self.compress = Tag.NO_COMPRESS.value not in tags



from __future__ import annotations

import logging
from enum import Enum
from typing import Literal

from .question import Question, Tag
from .repository import Repository

logger = logging.getLogger(__name__)
Mode = Literal["check", "interactive", "fix"]


class UserInput(Enum):
    REPLACE = "REPLACE"
    IGNORE = "IGNORE"
    MOVE_ON = "MOVE ON"
    REVIEW = "REVIEW"


def get_user_input() -> UserInput:
    print()
    response = input("Enter 'r' to replace, 'i' to ignore in future, 'v' to tag for review and move on, anything else to just move on: ")
    if response == "r":
        return UserInput("REPLACE")
    elif response == "i":
        return UserInput("IGNORE")
    elif response == "v":
        return UserInput("REVIEW")
    else:
        return UserInput("MOVE ON")


def check_output(repository: Repository, mode: Mode) -> int:
    failed: list[Question] = []
    fixed: list[Question] = []
    ignored: list[Question] = []
    questions = repository.get()
    questions_to_check = [question for question in questions if question.check_output]

    print(f"Found {len(questions)} questions.")
    print(f"Will check {len(questions_to_check)}.")
    print("----------")

    for question in questions_to_check:
        logger.info(f"Checking {question.id} output")
        normalised_actual_output = question.normalised_actual_output()
        if normalised_actual_output != question.given_output:
            print(f"\N{CROSS MARK} Bad output for {question.id}.", end="\n\n")
            print("Code:")
            colour_print(question.snippet.code, colour="cyan", end="\n\n")
            print("Output (normalised):")
            colour_print(normalised_actual_output, colour="green", end="\n\n")
            print("Given:")
            colour_print(question.given_output, colour="red")
            failed.append(question)
            if mode == "interactive":
                response = get_user_input()
                if response == UserInput.REPLACE:
                    repository.write_output(question, normalised_actual_output)
                    fixed.append(question)
                    print("\N{SPARKLES} Replaced.")
                elif response == UserInput.IGNORE:
                    repository.add_tag(question, Tag.NO_CHECK_OUTPUT)
                    ignored.append(question)
                    print("\N{SEE-NO-EVIL MONKEY} Will ignore in future.")
                elif response == UserInput.REVIEW:
                    repository.add_tag(question, Tag.REVIEW)
                    print("\N{RIGHT-POINTING MAGNIFYING GLASS} Added 'review' tag.")
                else:
                    print("\N{FACE WITHOUT MOUTH} Moving on.")
            elif mode == "fix":
                repository.write_output(question, normalised_actual_output)
                fixed.append(question)
                print("\N{SPARKLES} Auto-fixed.")
            print("----------")

    if failed:
        print(
            f"{len(failed)} questions had bad output "
            "("
            f"{len(fixed)} fixed, "
            f"{len(ignored)} will be ignored in future, "
            f"{len(failed) - len(fixed) - len(ignored)} remaining"
            ")"
        )
        return 1
    else:
        print("\N{WHITE HEAVY CHECK MARK} All good.")
        return 0


def check_formatting(repository: Repository, mode: Mode) -> int:
    failed: list[Question] = []
    fixed: list[Question] = []
    ignored: list[Question] = []
    questions = repository.get()
    questions_to_check = [question for question in questions if question.check_format]

    print(f"Found {len(questions)} questions.")
    print(f"Will check {len(questions_to_check)}.")
    print("----------")

    for question in questions_to_check:
        logger.info(f"Checking {question.id} format")
        formatted = question.snippet.format(compress=question.compress)
        if formatted is None:
            # error when trying to format snippet
            # treat as non-fixable failure
            print(f"\N{CROSS MARK} Error when formatting {question.id}.", end="\n\n")
            print("Given:")
            colour_print(question.snippet.code, colour="red")
            failed.append(question)
            print("----------")
        elif formatted != question.snippet.code:
            print(f"\N{CROSS MARK} Bad formatting for {question.id}.", end="\n\n")
            print("Formatted:")
            colour_print(formatted, colour="green", end="\n\n")
            print("Given:")
            colour_print(question.snippet.code, colour="red")
            failed.append(question)
            if mode == "interactive":
                response = get_user_input()
                if response == UserInput.REPLACE:
                    repository.write_code(question, formatted)
                    fixed.append(question)
                    print("\N{SPARKLES} Replaced.")
                elif response == UserInput.IGNORE:
                    repository.add_tag(question, Tag.NO_CHECK_FORMAT)
                    ignored.append(question)
                    print("\N{SEE-NO-EVIL MONKEY} Will ignore in future.")
                elif response == UserInput.REVIEW:
                    repository.add_tag(question, Tag.REVIEW)
                    print("\N{RIGHT-POINTING MAGNIFYING GLASS} Added 'review' tag.")
                else:
                    print("\N{FACE WITHOUT MOUTH} Moving on.")
            elif mode == "fix":
                repository.write_code(question, formatted)
                fixed.append(question)
                print("\N{SPARKLES} Auto-fixed.")
            print("----------")

    if failed:
        print(
            f"{len(failed)} questions had bad formatting "
            "("
            f"{len(fixed)} fixed, "
            f"{len(ignored)} will be ignored in future, "
            f"{len(failed) - len(fixed) - len(ignored)} left"
            ")"
        )
        return 1
    else:
        print("\N{WHITE HEAVY CHECK MARK} All good.")
        return 0


def colour_print(string: str, colour: str, **kwargs) -> None:
    if colour == "green":
        print("\033[92m" + string + "\033[0m", **kwargs)
    elif colour == "red":
        print("\033[91m" + string + "\033[0m", **kwargs)
    elif colour == "cyan":
        print("\033[96m" + string + "\033[0m", **kwargs)
    else:
        raise ValueError(f"unsupported colour: {colour}")



from enum import Enum
from pathlib import Path

from .normaliser import GoOutputNormaliser, NodeOutputNormaliser, OutputNormaliser, PythonOutputNormaliser, RubyOutputNormaliser
from .snippet import GoSnippet, NodeSnippet, PythonSnippet, RubySnippet, RustSnippet, Snippet


class Tag(Enum):
    "Question label, used to signal special treatment."

    NO_CHECK_FORMAT = "no_check_format"
    NO_CHECK_OUTPUT = "no_check_output"
    NO_COMPRESS = "no_compress"
    REVIEW = "review"


class Question:
    def __init__(
        self,
        id: int | Path,
        code: str,
        image: str,
        given_output: str,
        check_output: bool,
        check_format: bool,
        output_verbosity: int,
        compress: bool,
        review: bool = False,
    ):
        self.id = id
        self.image = image
        self.given_output = given_output
        self.output_verbosity = output_verbosity
        self.compress = compress
        self.snippet: Snippet
        self.output_normaliser: type[OutputNormaliser]
        if image.startswith("golang"):
            self.snippet = GoSnippet(code, image)
            self.output_normaliser = GoOutputNormaliser
        elif image.startswith("python") or image.startswith("numpy"):
            self.snippet = PythonSnippet(code, image)
            self.output_normaliser = PythonOutputNormaliser
        elif image.startswith("node"):
            self.snippet = NodeSnippet(code, image)
            self.output_normaliser = NodeOutputNormaliser
        elif image.startswith("ruby"):
            self.snippet = RubySnippet(code, image)
            self.output_normaliser = RubyOutputNormaliser
        elif image.startswith("rust"):
            self.snippet = RustSnippet(code, image)
            self.output_normaliser = RubyOutputNormaliser
        else:
            raise ValueError(f"Cannot tell language from image '{image}'")
        self.check_output = check_output
        self.check_format = check_format
        self.review = review

    def normalised_actual_output(self):
        actual_output = self.snippet.output()
        return self.output_normaliser.normalise(actual_output, output_verbosity=self.output_verbosity)



from __future__ import annotations

import re
from abc import ABC, abstractmethod
from itertools import count


class OutputNormaliser(ABC):
    @abstractmethod
    def normalise(self, output: str, output_verbosity: int) -> str:
        raise NotImplementedError


class PythonOutputNormaliser(OutputNormaliser):
    memory_address = re.compile(r"\b0x[0-9A-Fa-f]+\b")

    traceback_except_for_last_line = re.compile(
        r"Traceback \(most\ recent\ call\ last\):\n"  # start of traceback
        r"(\s.*\n)+",  # one or more lines starting with unicode whitespace and ending with newline
    )  # a traceback's last line doesn't start with whitespace so won't be captured

    # e.g. syntax errors include location info
    location_info = re.compile(
        r'  File "[^"]*", line.*\n'  # file and line number
        r".*\n"  # line
        r".*\n"  # carets
    )

    @classmethod
    def normalise(cls, output: str, output_verbosity: int) -> str:
        normalised = cls.normalise_memory_addresses(output)
        normalised = cls.normalise_traceback(normalised, output_verbosity)
        normalised = cls.normalise_location_info(normalised, output_verbosity)

        return normalised

    @classmethod
    def normalise_memory_addresses(cls, output: str) -> str:
        addresses = (hex(i) for i in count(0x100, 0x100))  # nice-looking, easily distinguished fake memory addresses
        seen = set()
        normalised = output
        for match in re.finditer(cls.memory_address, output):
            address = match.group()
            if address in seen:
                continue
            seen.add(address)
            normalised = normalised.replace(address, next(addresses))

        return normalised

    @classmethod
    def normalise_traceback(cls, output: str, output_verbosity: int) -> str:
        if output_verbosity == 0:
            normalised = re.sub(cls.traceback_except_for_last_line, "", output)
        elif output_verbosity == 1:
            normalised = re.sub(cls.traceback_except_for_last_line, "Traceback (most recent call last):\n  ...\n", output)
        elif output_verbosity == 2:
            normalised = output

        return normalised

    @classmethod
    def normalise_location_info(cls, output: str, output_verbosity: int) -> str:
        normalised = re.sub(cls.location_info, "", output) if output_verbosity < 2 else output
        return normalised


class GoOutputNormaliser(OutputNormaliser):
    memory_address = re.compile(r"\b0x[0-9A-Fa-f]+\b")
    panic = re.compile(
        r"(panic: .*?\n)"  # the line we want
        r".*",  # the rest
        re.DOTALL,
    )
    stack_overflow = re.compile(
        r"runtime: goroutine stack exceeds.*limit\n"
        r"runtime:.*\n"
        r"(fatal error: stack overflow\n)"  # the line we want
        r".*",  # the rest
        re.DOTALL,
    )

    @classmethod
    def normalise_memory_addresses(cls, output: str) -> str:
        addresses = (hex(i) for i in count(0x100, 0x100))  # nice-looking, easily distinguished fake memory addresses
        seen = set()
        normalised = output
        for match in re.finditer(cls.memory_address, output):
            address = match.group()
            if address in seen:
                continue
            seen.add(address)
            normalised = normalised.replace(address, next(addresses))

        return normalised

    @classmethod
    def normalise(cls, output: str, output_verbosity: int) -> str:
        normalised = cls.normalise_memory_addresses(output)
        normalised = cls.normalise_panic(normalised)
        normalised = cls.normalise_stack_overflow(normalised)

        return normalised

    @classmethod
    def normalise_panic(cls, output: str) -> str:
        normalised = re.sub(cls.panic, r"\1", output)
        return normalised

    @classmethod
    def normalise_stack_overflow(cls, output: str) -> str:
        normalised = re.sub(cls.stack_overflow, r"\1", output)
        return normalised


class NodeOutputNormaliser(OutputNormaliser):
    traceback = re.compile(
        r"(?P<location>/tmp/main.js:\d+\n)"
        r"(?P<offending_line>.*\n)"
        r"(?P<pointer>.*\n)"
        r"\n"  # empty line
        r"(?P<key_line>.*\n)"
        r"(.*\n)*"  # stack trace and empty line
        r"(?P<version>Node.js v.*\n)"  # version
    )

    @classmethod
    def normalise(cls, output: str, output_verbosity: int) -> str:
        normalised = cls.normalise_traceback(output, output_verbosity)
        return normalised

    @classmethod
    def normalise_traceback(cls, output: str, output_verbosity: int) -> str:
        if output_verbosity == 0:
            normalised = re.sub(cls.traceback, r"\g<key_line>", output)
        elif output_verbosity == 1:
            pass
        elif output_verbosity == 2:
            normalised = output

        return normalised


class RubyOutputNormaliser(OutputNormaliser):
    @classmethod
    def normalise(cls, output: str, output_verbosity: int) -> str:
        return output  # TODO:


class RustOutput(OutputNormaliser):
    @classmethod
    def normalise(cls, output: str, output_verbosity: int) -> str:
        return output  # TODO:



from __future__ import annotations

import atexit
import io
import logging
import platform
import tarfile
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import ClassVar

import docker

logger = logging.getLogger(__name__)


class DockerExecutor:
    _client: ClassVar[docker.client.DockerClient] = (
        docker.from_env()
        if platform.system() == "Linux"
        else docker.from_env(environment={"DOCKER_HOST": f"unix://{Path.home()}/.docker/run/docker.sock"})
    )

    _container_pool: ClassVar[dict[str, docker.models.containers.Container]] = {}

    def _get_container(self, image: str) -> docker.models.containers.Container:
        """Get or create a long-running container for the given image."""
        if image not in self._container_pool:
            try:
                self._client.images.get(image)
            except docker.errors.ImageNotFound:
                logger.info(f"Pulling image {image}...")
                self._client.images.pull(image)

            logger.info(f"Creating container from image {image}...")

            self._container_pool[image] = self._client.containers.run(
                image,
                ["tail", "-f", "/dev/null"],
                detach=True,
            )
        return self._container_pool[image]

    def write(self, image: str, content: str, dest: Path) -> None:
        container = self._get_container(image)
        data = io.BytesIO()
        with tarfile.open(fileobj=data, mode="w") as tar:
            file_bytes = content.encode("utf-8")
            tarinfo = tarfile.TarInfo(name=dest.name)
            tarinfo.size = len(file_bytes)
            tarinfo.mtime = time.time()
            tar.addfile(tarinfo, io.BytesIO(file_bytes))
        data.seek(0)

        succeeded = container.put_archive(str(dest.parent), data)
        assert succeeded

    @classmethod
    def cleanup(self) -> None:
        """Remove all containers in the pool."""
        for container in self._container_pool.values():
            container.remove(force=True)
        self._container_pool.clear()

    def exec_run(
        self,
        image: str,
        command: list[str],
        environment: dict[str, str] | None = None,
        workdir: str | None = None,
    ) -> tuple[int, bytes]:
        container = self._get_container(image)
        return container.exec_run(command, environment=environment, workdir=workdir)

    def exec_run_timed(
        self,
        image: str,
        command: list[str],
        environment: dict[str, str] | None = None,
        workdir: str | None = None,
    ) -> str:
        "Run the given command in a container, returning a string representing the timed output."
        container = self._get_container(image)
        logs: list[tuple[float, bytes]] = []
        previous = time.perf_counter()  # must come after getting the container
        _, output_stream = container.exec_run(command, tty=True, stream=True, environment=environment, workdir=workdir)
        for chunk in output_stream:
            now = time.perf_counter()
            logs.append((now - previous, chunk))
            previous = now

        return to_string(logs)


def to_string(logs: list[tuple[float, bytes]]) -> str:
    "Return a string constructed from timed docker logs."
    output = b""
    for delta, char in logs:
        rounded_delta = round(delta)
        if rounded_delta == 0:
            output += char
        else:
            output += bytes(f"<~{rounded_delta}s>\n", "utf-8")
            output += char

    return output.decode("utf-8").replace("\r\n", "\n")


class Snippet(ABC):
    """Abstract base class for code snippets in some language."""

    def __init__(self, code: str, image: str):
        self.code = code
        self.image = image
        self.executor = DockerExecutor()

    @abstractmethod
    def output(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def format(self, compress: bool) -> str | None:
        raise NotImplementedError


class PythonSnippet(Snippet):
    "A Python code snippet."

    def output(self) -> str:
        dest = Path("/tmp/main.py")
        self.executor.write(self.image, self.code, dest)
        return self.executor.exec_run_timed(
            self.image,
            ["python", str(dest)],
            environment={
                "NO_COLOR": "true",
                "PYTHONWARNINGS": "ignore",
            },
        )

    def format(self, compress: bool) -> str | None:
        dest = Path("/tmp/main.py")
        self.executor.write(self.image, self.code, dest)
        exit_code, foo = self.executor.exec_run(self.image, ["/bin/sh", "-c", f"python -m pip install ruff && ruff format {dest}"])
        _, bytes_ = self.executor.exec_run(self.image, ["cat", str(dest)])
        if exit_code != 0:
            formatted = None
        else:
            formatted = bytes_.decode("utf-8")
            if compress:
                formatted = formatted.replace("\n\n\n", "\n\n")  # crude

        return formatted


class GoSnippet(Snippet):
    "A Go code snippet."

    def output(self) -> str:
        dest = Path("/tmp/main.go")
        self.executor.write(self.image, self.code, dest)
        self.executor.exec_run(
            self.image,
            ["go", "build", str(dest)],
            workdir="/tmp",
        )
        return self.executor.exec_run_timed(self.image, ["/tmp/main"])

    def format(self, compress: bool = False) -> str | None:
        dest = Path("/tmp/main.go")
        self.executor.write(self.image, self.code, dest)
        exit_code, _ = self.executor.exec_run(self.image, ["go", "fmt", str(dest)])
        if exit_code != 0:
            formatted = None
        else:
            _, output = self.executor.exec_run(self.image, ["cat", str(dest)])
            formatted = output.decode("utf-8")
            if compress:
                formatted = formatted.strip().replace("\n\n\n", "\n\n")

        return formatted


class NodeSnippet(Snippet):
    "A Node code snippet."

    def output(self) -> str:
        dest = Path("/tmp/main.js")
        self.executor.write(self.image, self.code, dest)
        return self.executor.exec_run_timed(self.image, ["node", str(dest)], environment={"NO_COLOR": "1"})

    def format(self, compress: bool = False) -> str | None:
        dest = Path("/tmp/main.js")
        self.executor.write(self.image, self.code, dest)
        exit_code, _ = self.executor.exec_run(self.image, ["/bin/sh", "-c", f"npx prettier --write {dest}"])
        _, bytes_ = self.executor.exec_run(self.image, ["cat", str(dest)])
        if exit_code != 0:
            formatted = None
        else:
            formatted = bytes_.decode("utf-8")
            if compress:
                formatted = formatted.replace("\n\n\n", "\n\n")  # crude

        return formatted


class RubySnippet(Snippet):
    "A Ruby code snippet."

    def output(self) -> str:
        dest = Path("/tmp/main.rb")
        self.executor.write(self.image, self.code, dest)
        return self.executor.exec_run_timed(self.image, ["ruby", str(dest)])

    def format(self, compress: bool = False) -> str | None:
        dest = Path("/tmp/main.rb")
        self.executor.write(self.image, self.code, dest)
        exit_code, _ = self.executor.exec_run(self.image, ["/bin/sh", "-c", f"gem install rubocop && rubocop -A {dest}"])
        _, bytes_ = self.executor.exec_run(self.image, ["cat", str(dest)])
        if exit_code != 0:
            formatted = None
        else:
            formatted = bytes_.decode("utf-8")
            if compress:
                formatted = formatted.replace("\n\n\n", "\n\n")  # crude

        return formatted


class RustSnippet(Snippet):
    "A Rust code snippet."

    def output(self) -> str:
        dest = Path("/tmp/main.rs")
        self.executor.write(self.image, self.code, dest)
        self.executor.exec_run(self.image, ["rustc", "main.rs"], workdir="/tmp")
        return self.executor.exec_run_timed(self.image, ["/tmp/main"])

    def format(self, compress: bool = False) -> str | None:
        dest = Path("/tmp/main.rs")
        self.executor.exec_run(self.image, ["rustup", "component", "add", "rustfmt"])
        self.executor.write(self.image, self.code, dest)
        exit_code, _ = self.executor.exec_run(self.image, ["rustfmt", str(dest)])
        _, bytes_ = self.executor.exec_run(self.image, ["cat", str(dest)])
        if exit_code != 0:
            formatted = None
        else:
            formatted = bytes_.decode("utf-8")
            if compress:
                formatted = formatted.replace("\n\n\n", "\n\n")  # crude

        return formatted


atexit.register(DockerExecutor.cleanup)
